
__all__ = ['ContextManager', 'WrapWithContextManager']


class ContextManager(object):
    name = 'ContextManager'

    def on_start(self, request):
        pass

    def on_success(self, request):
        pass

    def on_failure(self, request):
        pass

    def wrap_call(self, func):
        return func


class WrapWithContextManager(object):

    def __init__(self, context=None, skip_list=None):
        self.context = context
        if skip_list is None:
            skip_list = []
        self.skip_list = skip_list

    def __call__(self, f):
        def wrap(fnc, ctm):
            def g(*a, **b):
                try:
                    ctm.on_start(a[0])
                    output = ctm.wrap_call(fnc)(*a, **b)
                    ctm.on_success(a[0])
                    return output

                except:
                    ctm.on_failure(a[0])
                    raise

            return g

        def wrap_skip_context(fnc, ctm):
            def g(*a, **b):
                output = ctm.wrap_call(fnc)(*a, **b)
                return output

            return g

        if self.context:
            for context in reversed(self.context):
                if isinstance(context, ContextManager):
                    if context.name in self.skip_list:
                        f = wrap_skip_context(f, context)
                    else:
                        f = wrap(f, context)

        return f



""" Messages module """

from datetime import timedelta

# Kaira import
from kaira.signing import dumps as signing_dumps
from kaira.signing import loads as signing_loads


DEBUG = 10
INFO = 20
SUCCESS = 25
WARNING = 30
ERROR = 40

DEFAULT_TAGS = {
    DEBUG: 'debug',
    INFO: 'info',
    SUCCESS: 'success',
    WARNING: 'warning',
    ERROR: 'error',
}

DEFAULT_LEVELS = {
    'DEBUG': DEBUG,
    'INFO': INFO,
    'SUCCESS': SUCCESS,
    'WARNING': WARNING,
    'ERROR': ERROR,
}

LEVEL_TAGS = DEFAULT_LEVELS


class Message:
    """
    Represent an actual message that can be stored in any of the supported
    storage classes (typically session- or cookie-based) and rendered in a view
    or template.
    """

    def __init__(self, level, message, extra_tags=None):
        self.level = int(level)
        self.message = message
        self.extra_tags = extra_tags

    def _prepare(self):
        """
        Prepare the message for serialization by forcing the ``message``
        and ``extra_tags`` to str in case they are lazy translations.
        """
        self.message = str(self.message)
        self.extra_tags = str(self.extra_tags) if self.extra_tags is not None else None

    def __eq__(self, other):
        return isinstance(other, Message) and self.level == other.level and \
            self.message == other.message

    def __str__(self):
        return str(self.message)

    @property
    def tags(self):
        return ' '.join(tag for tag in [self.extra_tags, self.level_tag] if tag)

    @property
    def level_tag(self):
        return LEVEL_TAGS.get(self.level, '')


class BaseStorage:
    """
    This is the base backend for temporary message storage.
    This is not a complete class; to be a usable storage backend, it must be
    subclassed and the two methods ``_get`` and ``_store`` overridden.
    """

    def __init__(self, request, *args, **kwargs):
        self.request = request
        self._queued_messages = []
        self.used = False
        self.added_new = False
        super().__init__(*args, **kwargs)

    def __len__(self):
        return len(self._loaded_messages) + len(self._queued_messages)

    def __iter__(self):
        self.used = True
        if self._queued_messages:
            self._loaded_messages.extend(self._queued_messages)
            self._queued_messages = []
        return iter(self._loaded_messages)

    def __contains__(self, item):
        return item in self._loaded_messages or item in self._queued_messages

    @property
    def _loaded_messages(self):
        """
        Return a list of loaded messages, retrieving them first if they have
        not been loaded yet.
        """
        if not hasattr(self, '_loaded_data'):
            messages, all_retrieved = self._get()
            self._loaded_data = messages or []
        return self._loaded_data

    def _get(self, *args, **kwargs):
        """
        Retrieve a list of stored messages. Return a tuple of the messages
        and a flag indicating whether or not all the messages originally
        intended to be stored in this storage were, in fact, stored and
        retrieved; e.g., ``(messages, all_retrieved)``.
        **This method must be implemented by a subclass.**
        If it is possible to tell if the backend was not used (as opposed to
        just containing no messages) then ``None`` should be returned in
        place of ``messages``.
        """
        raise NotImplementedError('subclasses of BaseStorage must provide a _get() method')

    def _store(self, messages, response, *args, **kwargs):
        """
        Store a list of messages and return a list of any messages which could
        not be stored.
        One type of object must be able to be stored, ``Message``.
        **This method must be implemented by a subclass.**
        """
        raise NotImplementedError('subclasses of BaseStorage must provide a _store() method')

    def _prepare_messages(self, messages):
        """
        Prepare a list of messages for storage.
        """
        for message in messages:
            message._prepare()

    def update(self, response):
        """
        Store all unread messages.
        If the backend has yet to be iterated, store previously stored messages
        again. Otherwise, only store messages added after the last iteration.
        """
        self._prepare_messages(self._queued_messages)
        if self.used:
            return self._store(self._queued_messages, response)
        elif self.added_new:
            messages = self._loaded_messages + self._queued_messages
            return self._store(messages, response)

    def add(self, level, message, extra_tags=''):
        """
        Queue a message to be stored.
        The message is only queued if it contained something and its level is
        not less than the recording level (``self.level``).
        """
        if not message:
            return
        # Check that the message level is not less than the recording level.
        level = int(level)
        if level < self.level:
            return
        # Add the message.
        self.added_new = True
        message = Message(level, message, extra_tags=extra_tags)
        self._queued_messages.append(message)

    def _get_level(self):
        """
        Return the minimum recorded level.
        The default level is the ``MESSAGE_LEVEL`` setting. If this is
        not found, the ``INFO`` level is used.
        """
        if not hasattr(self, '_level'):
            self._level = INFO
        return self._level

    def _set_level(self, value=None):
        """
        Set a custom minimum recorded level.
        If set to ``None``, the default level will be used (see the
        ``_get_level`` method).
        """
        if value is None and hasattr(self, '_level'):
            del self._level
        else:
            self._level = int(value)

    level = property(_get_level, _set_level, _set_level)


class MessageStorage(BaseStorage):
    """
    Store messages in a cookie.
    """
    cookie_name = 'messages'
    # uwsgi's default configuration enforces a maximum size of 4kb for all the
    # HTTP headers. In order to leave some room for other cookies and headers,
    # restrict the session cookie to 1/2 of 4kb. See #18781.
    max_cookie_size = 2048
    not_finished = '__messagesnotfinished__'

    def __init__(self, request, *args, **kwargs):

        self.request = request
        self.cookies = kwargs.pop('cookies', None)
        options = kwargs.pop('options', None)
        if not options:
            options = {
                       'MESSAGES_COOKIE_DOMAIN': '',
                       'MESSAGES_COOKIE_SECURE': False,
                       'MESSAGES_COOKIE_HTTPONLY': True,
                       'MESSAGES_COOKIE_EXPIRE': 600,
                       'MESSAGES_COOKIE_PATH': '/',
                       'MESSAGES_COOKIE_NAME': 'kaira_message',
                       'MESSAGES_SECRET': b'APj00cpfa8Gx1SjnyLxwBBSQfnQ9DJYe0Cm',
                       'MESSAGES_TIME_LIMIT': timedelta(minutes=60)
                      }
        self.options = options
        self._queued_messages = []
        self.used = False
        self.added_new = False
        super().__init__(*args, **kwargs)

    def _get(self, *args, **kwargs):
        """
        Retrieve a list of messages from the messages cookie. If the
        not_finished sentinel value is found at the end of the message list,
        remove it and return a result indicating that not all messages were
        retrieved by this storage.
        """
        cookie_name = self.options['MESSAGES_COOKIE_NAME']
        data = self.request.cookies[cookie_name]
        messages = self._decode(data)
        all_retrieved = not (messages and messages[-1] == self.not_finished)
        if messages and not all_retrieved:
            # remove the sentinel value
            messages.pop()
        return messages, all_retrieved

    def _update_cookie(self, encoded_data, response):
        """
        Either set the cookie with the encoded data if there is any data to
        store, or delete the cookie.
        """
        if encoded_data:
            response.set_cookie(
                self.cookie_name, encoded_data,
                domain=settings.SESSION_COOKIE_DOMAIN,
                secure=settings.SESSION_COOKIE_SECURE or None,
                httponly=settings.SESSION_COOKIE_HTTPONLY or None,
                samesite=settings.SESSION_COOKIE_SAMESITE,
            )
        else:
            response.delete_cookie(self.cookie_name, domain=settings.SESSION_COOKIE_DOMAIN)

    def _store(self, messages, response, remove_oldest=True, *args, **kwargs):
        """
        Store the messages to a cookie and return a list of any messages which
        could not be stored.
        If the encoded data is larger than ``max_cookie_size``, remove
        messages until the data fits (these are the messages which are
        returned), and add the not_finished sentinel value to indicate as much.
        """
        unstored_messages = []
        encoded_data = self._encode(messages)
        if self.max_cookie_size:
            # data is going to be stored eventually by SimpleCookie, which
            # adds its own overhead, which we must account for.
            cookie = SimpleCookie()  # create outside the loop

            def stored_length(val):
                return len(cookie.value_encode(val)[1])

            while encoded_data and stored_length(encoded_data) > self.max_cookie_size:
                if remove_oldest:
                    unstored_messages.append(messages.pop(0))
                else:
                    unstored_messages.insert(0, messages.pop())
                encoded_data = self._encode(messages + [self.not_finished],
                                            encode_empty=unstored_messages)
        self._update_cookie(encoded_data, response)
        return unstored_messages

    def _encode(self, messages, encode_empty=False):
        """
        Return an encoded version of the messages list which can be stored as
        plain text.
        Since the data will be retrieved from the client-side, the encoded data
        also contains a hash to ensure that the data was not tampered with.
        """
        if messages or encode_empty:
            value = signing_dumps(messages,
                                  key=self.options['MESSAGES_SECRET'])
            return value

    def _decode(self, data):
        """
        Safely decode an encoded text stream back into a list of messages.
        If the encoded text stream contained an invalid hash or was in an
        invalid format, return None.
        """
        if not data:
            return None

        try:
            messages = signing_loads(data,
                                     key=self.options['MESSAGES_SECRET'],
                                     max_age=self.options['MESSAGES_TIME_LIMIT'])
        except:
            messages = None
            self.used = True

        return messages


# import datetime
# from datetime import timedelta
#
#
# class BaseFlashManager(object):
#     """Flash Manager"""
#
#     def __init__(self, request, msg, status, cookies=None, options=None):
#         """ Init """
#
#         self.request = request
#         self.cookies = cookies
#         self.msg = msg
#         self.status = status
#
#         if not options:
#             options = {
#                 'FLASH_COOKIE_DOMAIN': '',
#                 'FLASH_COOKIE_SECURE': False,
#                 'FLASH_COOKIE_HTTPONLY': True,
#                 'FLASH_COOKIE_EXPIRE': 600,
#                 'FLASH_COOKIE_PATH': '/',
#                 'FLASH_COOKIE_NAME': 'kaira_flash',
#                 'FLASH_SECRET': b'APj00cpfa8Gx1SjnyLxwBBSQfnQ9DJYe0Cm',
#                 'FLASH_TIME_LIMIT': timedelta(minutes=60)
#             }
#
#         self.options = options
#
#     def save(self):
#         """Save"""
#
#         expire_seconds = int(self.options['FLASH_COOKIE_EXPIRE'])
#         if expire_seconds > 0:
#             now = datetime.datetime.utcnow()
#             expires_cookie = now + datetime.timedelta(seconds=expire_seconds)
#         else:
#             expires_cookie = None
#
#         if not self.options['FLASH_COOKIE_DOMAIN'] or self.options['FLASH_COOKIE_DOMAIN'] == "":
#             domain = self.request.host.split(':')[0]
#         else:
#             domain = self.options['FLASH_COOKIE_DOMAIN']
#
#         cookies_options = {
#             'HTTP_COOKIE_DOMAIN': domain,
#             'HTTP_COOKIE_SECURE': self.options['FLASH_COOKIE_SECURE'],
#             'HTTP_COOKIE_HTTPONLY': self.options['FLASH_COOKIE_HTTPONLY']
#         }
#
#         cookie_name = self.options['FLASH_COOKIE_NAME']
#         time_limit = meta.csrf_options['FLASH_TIME_LIMIT']
#         csrf_secret = meta.csrf_options['CSRF_SECRET']
#
#         rand_code = None
#         if request.cookies:
#             if cookie_name in request.cookies:
#                 rand_code = request.cookies[cookie_name]
#
#         if not rand_code:
#             rand_code = sha1(os.urandom(64)).hexdigest()
#
#         if time_limit:
#             expires = (self.now() + time_limit).strftime(self.TIME_FORMAT)
#             csrf_build = '%s%s' % (rand_code, expires)
#         else:
#             expires = ''
#             csrf_build = rand_code
#
#         hmac_csrf = hmac.new(csrf_secret, csrf_build.encode('utf8'), digestmod=sha1)
#         value_csrf = '%s##%s' % (expires, hmac_csrf.hexdigest())
#
#         cookies = meta.cookies
#         if not cookies:
#             cookies = CookieManager(options=cookies_options)
#
#         cookies[cookie_name] = rand_code
#         cookies[cookie_name].path = meta.csrf_options['CSRF_COOKIE_PATH']
#         cookies[cookie_name].expires = expires_cookie
#
#         meta.cookies = cookies




import re
import functools

from .utils import sanitize


__all__ = ['tag', 'cat', 'safe']


# web2py style helpers

def xmlescape(s, quote=True):
    """
    returns an escaped string of the provided text s
    s: the text to be escaped
    quote: optional (default True)
    """
    # first try the xml function
    if isinstance(s, TAG):
        return s.xml()
    # otherwise, make it a string
    if isinstance(s, bytes):
        s = s.encode('utf-8')
    s = s.replace("&", "&amp;") # Must be done first!
    s = s.replace("<", "&lt;")
    s = s.replace(">", "&gt;")
    if quote:
        s = s.replace("'", "&#x27;")
        s = s.replace('"', "&quot;")
    return s


class TAG(object):

    rules = {'ul':['li'],
             'ol':['li'],
             'table':['tr','thead','tbody'],
             'thead':['tr'],
             'tbody':['tr'],
             'tr':['td','th'],
             'select':['option','optgroup'],
             'optgroup':['optionp']}

    def __init__(self, name):
        self.safe = safe
        self.name = name
        self.parent = None
        self.components = []
        self.attributes = {}

    @staticmethod
    def wrap(component,rules):
        if not rules:
            return component
        if not isinstance(component,TAG) or not component.name in rules:
            return TAG(rules[0])(component)

    def __call__(self, *components, **attributes):
        rules = self.rules.get(self.name,[])
        self.components = [self.wrap(comp,rules) for comp in components]
        self.attributes = attributes
        for component in self.components:
            if isinstance(component,TAG):
                component.parent = self
        return self

    def append(self,component):
        self.components.append(component)

    def insert(self,i,component):
        self.components.insert(i,component)

    def remove(self,component):
        self.components.remove(component)

    def __getitem__(self,key):
        if isinstance(key,int):
            return self.components[key]
        else:
            return self.attributes.get(key)

    def __setitem__(self,key,value):
        if isinstance(key,int):
            self.components.insert(key,value)
        else:
            self.attributes[key] = value

    def __iter__(self):
        for item in self.components:
            yield item

    def __str__(self):
        return self.xml()

    def __add__(self,other):
        return cat(self,other)

    def add_class(self, name):
        """ add a class to _class attribute """
        c = self['_class']
        classes = (set(c.split()) if c else set()) | set(name.split())
        self['_class'] = ' '.join(classes) if classes else None
        return self

    def remove_class(self, name):
        """ remove a class from _class attribute """
        c = self['_class']
        classes = (set(c.split()) if c else set()) - set(name.split())
        self['_class'] = ' '.join(classes) if classes else None
        return self

    regex_tag = re.compile('^([\w\-\:]+)')
    regex_id = re.compile('#([\w\-]+)')
    regex_class = re.compile('\.([\w\-]+)')
    regex_attr = re.compile('\[([\w\-\:]+)=(.*?)\]')

    def find(self,expr):
        union = lambda a,b: a.union(b)
        if ',' in expr:
            tags = functools.reduce(union, [self.find(x.strip())
                                  for x in expr.split(',')],set())
        elif ' ' in expr:
            tags = [self]
            for k,item in enumerate(expr.split()):
                if k>0:
                    children = [set([c for c in tag if isinstance(c,TAG)])
                                for tag in tags]
                    tags = functools.reduce(union,children)
                tags = functools.reduce(union, [tag.find(item) for tag in tags],set())
        else:
            tags = functools.reduce(union,[c.find(expr)
                                 for c in self if isinstance(c,TAG)],set())
            tag = TAG.regex_tag.match(expr)
            id = TAG.regex_id.match(expr)
            _class = TAG.regex_class.match(expr)
            attr = TAG.regex_attr.match(expr)
            if (tag is None or self.name == tag.group(1)) and \
               (id is None or self['_id'] == id.group(1)) and \
               (_class is None or _class.group(1) in \
                    (self['_class'] or '').split()) and \
               (attr is None or self['_'+attr.group(1)] == attr.group(2)):
                tags.add(self)
        return tags

    def xml(self):
        name = self.name
        co = ''.join(xmlescape(v) for v in self.components)
        ca = ' '.join('%s="%s"' % (k[1:], k[1:] if v==True else xmlescape(v))
             for (k, v) in sorted(self.attributes.items())
             if k.startswith('_') and v is not None)
        ca = ' ' + ca if ca else ''
        if not self.components:
            return '<%s%s />' % (name, ca)
        else:
            return '<%s%s>%s</%s>' % (name, ca, co, name)

    __repr__ = __str__


class METATAG(object):

    def __getattr__(self, name):
        return TAG(name)

    def __getitem__(self, name):
        return TAG(name)

tag = METATAG()


class cat(TAG):

    def __init__(self, *components):
        self.components = components

    def xml(self):
        return ''.join(xmlescape(v) for v in self.components)


class safe(TAG):

    default_allowed_tags = {
        'a':['href','title','target'], 'b':[], 'blockquote':['type'],
        'br':[], 'i':[], 'li':[], 'ol':[], 'ul':[], 'p':[], 'cite':[],
        'code':[], 'pre':[], 'img':['src', 'alt'], 'strong':[],
        'h1':[], 'h2':[], 'h3':[], 'h4':[], 'h5':[], 'h6':[],
        'table':[], 'tr':[], 'td':['colspan'], 'div':[],
        }

    def __init__(self, text, sanitize=False, allowed_tags=None):
        self.text = text
        self.allowed_tags = allowed_tags or safe.default_allowed_tags

    def xml(self):
        return self.text if not sanitize else \
            sanitize(self.text, self.allowed_tags.keys(), self.allowed_tags)



import base64
import pickle
from datetime import timedelta
import datetime
import hmac
from hashlib import sha1
import os
import hashlib

from wtforms import Form
from wtforms.widgets import HiddenInput
from wtforms.csrf.core import CSRF
from wtforms import ValidationError

from multidict import CIMultiDict

from kaira.cookie import CookieManager


SUBMIT_METHODS = ('POST', 'PUT', 'PATCH', 'DELETE')


class KairaFormBase(Form):

    def __init__(self, request, **kwargs):

        self.request = request

        self.form_name = 'kaira_form_'
        input_vars = dict()
        if request.method == 'GET':
            input_method = request.query
        elif request.method == 'POST':
            input_method = request.form
        for key, value in input_method.items():
            if not key:
                continue
            if isinstance(value, list) and len(value) == 1:
                value = value[0]
            input_vars[key] = value

        if input_vars:
            input_vars = CIMultiDict(**input_vars)
        else:
            input_vars = None

        super().__init__(formdata=input_vars, **kwargs)
        #self.generate_form_name()  # Generamos el nombre del form (uuid)

    def is_submitted(self):
        """Consider the form submitted if there is an active request and
        the method is ``POST``, ``PUT``, ``PATCH``, or ``DELETE``.
        """

        return bool(self.request) and self.request.method in SUBMIT_METHODS

    def validate_on_submit(self):
        """Call :meth:`validate` only if the form is submitted.
        This is a shortcut for ``form.is_submitted() and form.validate()``.
        """
        return self.is_submitted() and self.validate()

    def hidden_tag(self, *fields):
        """Render the form's hidden fields in one call.
        A field is considered hidden if it uses the
        :class:`~wtforms.widgets.HiddenInput` widget.
        If ``fields`` are given, only render the given fields that
        are hidden.  If a string is passed, render the field with that
        name if it exists.
        .. versionchanged:: 0.13
           No longer wraps inputs in hidden div.
           This is valid HTML 5.
        .. versionchanged:: 0.13
           Skip passed fields that aren't hidden.
           Skip passed names that don't exist.
        """

        def hidden_fields(fields):
            for f in fields:
                if isinstance(f, str):
                    f = getattr(self, f, None)

                if f is None or not isinstance(f.widget, HiddenInput):
                    continue

                yield f

        return u'\n'.join(str(f) for f in hidden_fields(fields or self))

    def generate_form_name(self):
        """Generamos el id"""

        l_fields = list()
        for f in self:
            l_fields.append(f.id)
        sums_ids = ''.join(l_fields)
        hash_code = hashlib.sha256(sums_ids.encode('utf-8')).hexdigest()

        self.form_name = 'kaira_form_' + hash_code

    def render_boostrap_fields(self):
        """ Render fields"""

        xml_form = list()
        for field in self:
            if field.id == 'csrf_token':  # salteamos el csrf token
                continue
            xml_field = list()
            xml_field.append('%s' % field.label)
            xml_field.append('%s' % field(class_='form-control'))
            if field.errors:
                xml_error = list()
                for error in field.errors:
                    xml_error.append('<li>%s</li>' % error)
                xml_field.append('<small class="form-text text-muted">%s</small>' % '\n'.join(xml_error))
            xml_form.append('<div class="form-group">%s</div>' % '\n'.join(xml_field))
        return '\n'.join(xml_form)

    def render_boostrap_form(self, path='/'):
        """ Render form"""

        r_fields = self.render_boostrap_fields()
        submit = '<button type="submit"> Send </button>'
        xml = '<form method="POST" enctype="multipart/form-data" action="{path}">{fields} {submit}</form>'\
            .format(path=path, fields=r_fields, submit=submit)
        return xml


class KSessionCSRF(CSRF):
    TIME_FORMAT = '%Y%m%d%H%M%S'

    def setup_form(self, form):

        self.form_meta = form.meta
        return super(KSessionCSRF, self).setup_form(form)

    def generate_csrf_token(self, csrf_token_field):

        meta = self.form_meta
        request = meta.request

        expire_seconds = int(meta.csrf_options['CSRF_COOKIE_EXPIRE'])
        if expire_seconds > 0:
            now = datetime.datetime.utcnow()
            expires_cookie = now + datetime.timedelta(seconds=expire_seconds)
        else:
            expires_cookie = None

        if not meta.csrf_options['CSRF_COOKIE_DOMAIN'] or meta.csrf_options['CSRF_COOKIE_DOMAIN'] == "":
            domain = meta.request.host.split(':')[0]
        else:
            domain = meta.csrf_options['CSRF_COOKIE_DOMAIN']

        cookies_options = {
            'HTTP_COOKIE_DOMAIN': domain,
            'HTTP_COOKIE_SECURE': meta.csrf_options['CSRF_COOKIE_SECURE'],
            'HTTP_COOKIE_HTTPONLY': meta.csrf_options['CSRF_COOKIE_HTTPONLY']
        }

        cookie_name = meta.csrf_options['CSRF_COOKIE_NAME'] + "_" + meta.form_name
        time_limit = meta.csrf_options['CSRF_TIME_LIMIT']
        csrf_secret = meta.csrf_options['CSRF_SECRET']

        rand_code = None
        if request.cookies:
            if cookie_name in request.cookies:
                rand_code = request.cookies[cookie_name]

        if not rand_code:
            rand_code = sha1(os.urandom(64)).hexdigest()

        if time_limit:
            expires = (self.now() + time_limit).strftime(self.TIME_FORMAT)
            csrf_build = '%s%s' % (rand_code, expires)
        else:
            expires = ''
            csrf_build = rand_code

        hmac_csrf = hmac.new(csrf_secret, csrf_build.encode('utf8'), digestmod=sha1)
        value_csrf = '%s##%s' % (expires, hmac_csrf.hexdigest())

        cookies = meta.cookies
        if not cookies:
            cookies = CookieManager(options=cookies_options)

        cookies[cookie_name] = rand_code
        cookies[cookie_name].path = meta.csrf_options['CSRF_COOKIE_PATH']
        cookies[cookie_name].expires = expires_cookie

        meta.cookies = cookies

        return value_csrf

    def delete_session(self):

        meta = self.form_meta
        cookie_name = meta.csrf_options['CSRF_COOKIE_NAME'] + "_" + meta.form_name

        if not meta.csrf_options['CSRF_COOKIE_DOMAIN'] or meta.csrf_options['CSRF_COOKIE_DOMAIN'] == "":
            domain = meta.request.host.split(':')[0]
        else:
            domain = meta.csrf_options['CSRF_COOKIE_DOMAIN']

        cookies_options = {
            'HTTP_COOKIE_DOMAIN': domain,
            'HTTP_COOKIE_SECURE': meta.csrf_options['CSRF_COOKIE_SECURE'],
            'HTTP_COOKIE_HTTPONLY': meta.csrf_options['CSRF_COOKIE_HTTPONLY']
        }

        cookies = meta.cookies
        if not cookies:
            cookies = CookieManager(options=cookies_options)

        del cookies[cookie_name]

        meta.cookies = cookies

    def validate_csrf_token(self, form, field):

        meta = self.form_meta

        time_limit = meta.csrf_options['CSRF_TIME_LIMIT']
        csrf_secret = meta.csrf_options['CSRF_SECRET']
        cookie_name = meta.csrf_options['CSRF_COOKIE_NAME'] + "_" + meta.form_name
        request = meta.request

        if not field.data or '##' not in field.data:
            raise ValidationError(field.gettext('CSRF token missing'))

        expires, hmac_csrf = field.data.split('##', 1)

        check_val = (str(request.cookies[cookie_name]) + expires).encode('utf8')

        hmac_compare = hmac.new(csrf_secret, check_val, digestmod=sha1)
        if hmac_compare.hexdigest() != hmac_csrf:
            raise ValidationError(field.gettext('CSRF failed'))

        if time_limit:
            now_formatted = self.now().strftime(self.TIME_FORMAT)
            if now_formatted > expires:
                raise ValidationError(field.gettext('CSRF token expired'))

    def now(self):
        """
        Get the current time. Used for test mocking/overriding mainly.
        """
        return datetime.datetime.now()


class KairaForm(KairaFormBase):

    class Meta:
        csrf = True
        csrf_class = KSessionCSRF
        csrf_options = {
            'CSRF_COOKIE_DOMAIN': '',
            'CSRF_COOKIE_SECURE': False,
            'CSRF_COOKIE_HTTPONLY': True,
            'CSRF_COOKIE_EXPIRE': 3600,
            'CSRF_COOKIE_PATH': '/',
            'CSRF_COOKIE_NAME': 'csrf',
            'CSRF_SECRET': b'APj00jpfj8Gx1SjnyLxwBBSQfnQ9DJYe0Cm',
            'CSRF_TIME_LIMIT': timedelta(minutes=60)
            }

    def __init__(self, request, cookies=None, csrf_options=None, form_name='my_form', **kwargs):

        self.request = request
        self.csrf_options = csrf_options
        self.cookies = cookies
        if not csrf_options:
            csrf_options = {
                'CSRF_COOKIE_DOMAIN': '',
                'CSRF_COOKIE_SECURE': False,
                'CSRF_COOKIE_HTTPONLY': True,
                'CSRF_COOKIE_EXPIRE': 3600,
                'CSRF_COOKIE_PATH': '/',
                'CSRF_COOKIE_NAME': 'csrf',
                'CSRF_SECRET': b'APj00jpfj8Gx1SjnyLxwBBSQfnQ9DJYe0Cm',
                'CSRF_TIME_LIMIT': timedelta(minutes=60)
            }

        super().__init__(request=request, meta=dict(cookies=cookies,
                                                    request=request,
                                                    csrf_options=csrf_options,
                                                    form_name=form_name), **kwargs)

    def render_boostrap_form(self, path=None, send_caption='Send'):
        """ Render form"""

        r_fields = self.render_boostrap_fields()
        if self.csrf_token.errors:
            csrf_errors = self.csrf_token.errors
        else:
            csrf_errors = ''
        if not path:
            path = self.request.path
        submit = '<button type="submit"> {send_caption} </button>'.format(send_caption=send_caption)
        xml = '<form method="POST" enctype="multipart/form-data" action="{path}">' \
              '{csrf_token} {csrf_errors}' \
              '{fields} ' \
              '{submit}' \
              '</form>'\
            .format(path=path,
                    fields=r_fields,
                    submit=submit,
                    csrf_token=self.csrf_token,
                    csrf_errors=csrf_errors)
        return xml


def read_in_chunks(file_object, chunk_size=1024):
    """Lazy function (generator) to read a file piece by piece.
    Default chunk size: 1k."""
    while True:
        data = file_object.read(chunk_size)
        if not data:
            break
        yield data


def upload_store(file_dest, upload_file):
    """Store the file"""

    upload_file.file.seek(0)
    for piece in read_in_chunks(upload_file.file):
        file_dest.write(piece)

    return file_dest


def serialize(obj):
    return base64.b64encode(pickle.dumps(obj))


def deserialize(s):
    return pickle.loads(base64.b64decode(s))




""" ``response`` module.
"""
import functools

try:
    from ujson import dumps as json_dumps
except:
    from json import dumps as json_dumps

from .templates import MakoTemplate, template
from kaira.wrapper import ContextManager
from kaira.messages import MessagesManager
from kaira.cookie import CookieManager

# see http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
# see http://en.wikipedia.org/wiki/List_of_HTTP_status_codes
HTTP_STATUS = {
    # Informational
    100: '100 Continue',
    101: '101 Switching Protocols',
    # Successful
    200: '200 OK',
    201: '201 Created',
    202: '202 Accepted',
    203: '203 Non-Authoritative Information',
    204: '204 No Content',
    205: '205 Reset Content',
    206: '206 Partial Content',
    207: '207 Multi-Status',
    # Redirection
    300: '300 Multiple Choices',
    301: '301 Moved Permanently',
    302: '302 Found',
    303: '303 See Other',
    304: '304 Not Modified',
    305: '305 Use Proxy',
    307: '307 Temporary Redirect',
    # Client Error
    400: '400 Bad Request',
    401: '401 Unauthorized',
    402: '402 Payment Required',
    403: '403 Forbidden',
    404: '404 Not Found',
    405: '405 Method Not Allowed',
    406: '406 Not Acceptable',
    407: '407 Proxy Authentication Required',
    408: '408 Request Timeout',
    409: '409 Conflict',
    410: '410 Gone',
    411: '411 Length Required',
    412: '412 Precondition Failed',
    413: '413 Request Entity Too Large',
    414: '414 Request-Uri Too Long',
    415: '415 Unsupported Media Type',
    416: '416 Requested Range Not Satisfiable',
    417: '417 Expectation Failed',
    # Server Error
    500: '500 Internal Server Error',
    501: '501 Not Implemented',
    502: '502 Bad Gateway',
    503: '503 Service Unavailable',
    504: '504 Gateway Timeout',
    505: '505 Http Version Not Supported'
}

HTTP_HEADER_CACHE_CONTROL_DEFAULT = ('Cache-Control', 'private')


class BASEHTTPResponse(object):
    status_code = 200
    cache_policy = None
    cache_profile = None


class HTTPResponse(BASEHTTPResponse):
    """ HTTP response """

    def __init__(self,
                 content_type='text/html; charset=UTF-8',
                 encoding='UTF-8',
                 status_code=200):
        """ Initializes HTTP response.
        """

        self.status_code = status_code
        self.content_type = content_type
        self.encoding = encoding
        self.headers = [('Content-Type', content_type)]
        self.buffer = []
        self.cookies = []
        self.cache_dependency = []

    def get_status(self):
        """ Returns a string that describes the specified
            HTTP status code.
        """
        return HTTP_STATUS[self.status_code]

    status = property(get_status)

    def redirect(self, absolute_url, status_code=302):
        """ Redirect response to ``absolute_url`` and sets
            ``status_code``.
        """
        self.status_code = status_code
        self.headers.append(('Location', absolute_url))

    def write(self, chunk):
        """ Applies encoding to ``chunk`` and append it to response
            buffer.
        """
        self.buffer.append(chunk.encode(self.encoding))

    def write_bytes(self, chunk):
        """ Appends chunk it to response buffer. No special checks performed.
            It must be valid object for WSGI response.
        """
        self.buffer.append(chunk)

    def __call__(self, start_response):
        """ WSGI call processing."""
        headers = self.headers
        append = headers.append
        cache_policy = self.cache_policy
        if cache_policy:
            cache_policy.extend(headers)
        else:
            append(HTTP_HEADER_CACHE_CONTROL_DEFAULT)
        if self.cookies:
            encoding = self.encoding
            if isinstance(self.cookies, dict):
                for key, cookie in self.cookies.items():
                    append(cookie.http_set_cookie(encoding))
            else:
                for cookie in self.cookies:
                    append(cookie.http_set_cookie(encoding))
        buffer = self.buffer
        append(('Content-Length', str(sum([len(chunk) for chunk in buffer]))))
        start_response(HTTP_STATUS[self.status_code], headers)
        return buffer


class HandleResponse:

    def __init__(self, template_adapter=MakoTemplate):
        self.render_template = functools.partial(template, template_adapter=template_adapter)

    @staticmethod
    def text(body, status_code=200, cookies=None, encoding="utf-8", headers=None):
        """ text """

        resp = HTTPResponse('text/plain; charset=' + encoding, encoding,
                            status_code=status_code)
        if cookies:
            resp.cookies = cookies
        if headers:
            for header in headers:
                resp.headers.append(header)

        resp.write_bytes(body.encode(encoding))

        return resp

    @staticmethod
    def json(body, status_code=200, cookies=None, encoding="utf-8", headers=None):
        """ json """

        resp = HTTPResponse('application/json; charset=' + encoding, encoding,
                            status_code=status_code)
        if cookies:
            resp.cookies = cookies
        if headers:
            for header in headers:
                resp.headers.append(header)

        resp.write_bytes(json_dumps(body).encode(encoding))

        return resp

    @staticmethod
    def html(body, status_code=200, cookies=None, encoding="utf-8", headers=None):
        """ html """

        resp = HTTPResponse('text/html; charset=' + encoding, encoding,
                            status_code=status_code)
        if cookies:
            resp.cookies = cookies
        if headers:
            for header in headers:
                resp.headers.append(header)

        resp.write_bytes(body.encode(encoding))

        return resp

    @staticmethod
    def redirect(to, status_code=302, cookies=None, encoding="utf-8", headers=None):
        """ redirect """

        resp = HTTPResponse('text/html; charset=' + encoding, encoding,
                            status_code=status_code)
        if cookies:
            resp.cookies = cookies
        if headers:
            for header in headers:
                resp.headers.append(header)

        resp.redirect(to, status_code=status_code)

        return resp

    @staticmethod
    def error(body, status_code=400, cookies=None, encoding="utf-8", headers=None):
        """ error """

        resp = HTTPResponse('text/html; charset=' + encoding, encoding,
                            status_code=status_code)

        if cookies:
            resp.cookies = cookies
        if headers:
            for header in headers:
                resp.headers.append(header)

        resp.write_bytes(body.encode(encoding))

        return resp

    def template(self, tpl, status_code=200, cookies=None, encoding="utf-8", headers=None, **kwargs):
        """ template """

        body = self.render_template(tpl, **kwargs)

        resp = HTTPResponse('text/html; charset=' + encoding, encoding,
                            status_code=status_code)
        if cookies:
            resp.cookies = cookies
        if headers:
            for header in headers:
                resp.headers.append(header)

        resp.write_bytes(body.encode(encoding))

        return resp


response = HandleResponse()


class ContextResponse(ContextManager):
    messages = None
    cookies = None

    def __init__(self, options=None):
        self.options = options
        self.messages = MessagesManager(options)
        self.cookies = None

    def on_start(self, request):
        self.messages.start(request)

    def template(self, tpl, status_code=200,  cookies=None, encoding="utf-8", headers=None, **kwargs):

        if self.messages.response_cookies:
            cookies = self.messages.response_cookies

        kwargs['flash_messages'] = self.messages

        return response.template(tpl, status_code=status_code, cookies=cookies, encoding=encoding, headers=headers, **kwargs)

    def error(self, body, status_code=400, cookies=None, encoding="utf-8", headers=None):

        return response.error(body, status_code=status_code, cookies=cookies, encoding=encoding, headers=headers)

    def redirect(self, to, status_code=302, cookies=None, encoding="utf-8", headers=None):
        """ redirect """

        return response.redirect(to, status_code=status_code, cookies=cookies, encoding=encoding, headers=headers)

    def html(self, body, status_code=200, cookies=None, encoding="utf-8", headers=None):

        return response.html(body, status_code=status_code, cookies=cookies, encoding=encoding, headers=headers)

    def json(self, body, status_code=200, cookies=None, encoding="utf-8", headers=None):

        return response.json(body, status_code=status_code, cookies=cookies, encoding=encoding, headers=headers)

    def text(self, body, status_code=200, cookies=None, encoding="utf-8", headers=None):

        return response.text(body, status_code=status_code, cookies=cookies, encoding=encoding, headers=headers)





class ServerAdapter(object):
    quiet = False

    def __init__(self, host='127.0.0.1', port=8080, **options):
        self.options = options
        self.host = host
        self.port = int(port)

    def run(self, handler):  # pragma: no cover
        pass

    def __repr__(self):
        args = ', '.join(['%s=%s' % (k, repr(v))
                          for k, v in self.options.items()])
        return "%s(%s)" % (self.__class__.__name__, args)


class WSGIRefServer(ServerAdapter):
    def run(self, app):  # pragma: no cover
        from wsgiref.simple_server import make_server
        from wsgiref.simple_server import WSGIRequestHandler, WSGIServer
        import socket

        class FixedHandler(WSGIRequestHandler):
            def address_string(self):  # Prevent reverse DNS lookups please.
                return self.client_address[0]

            def log_request(*args, **kw):
                if not self.quiet:
                    return WSGIRequestHandler.log_request(*args, **kw)

        handler_cls = FixedHandler
        server_cls = WSGIServer

        if ':' in self.host:  # Fix wsgiref for IPv6 addresses.
            if getattr(server_cls, 'address_family') == socket.AF_INET:

                class server_cls(server_cls):
                    address_family = socket.AF_INET6

        self.srv = make_server(self.host, self.port, app, server_cls,
                               handler_cls)
        self.port = self.srv.server_port  # update port actual port (0 means random)
        try:
            self.srv.serve_forever()
        except KeyboardInterrupt:
            self.srv.server_close()  # Prevent ResourceWarning: unclosed socket
            raise


server_names = {
    'wsgiref': WSGIRefServer
    }


def run(app=None, server='wsgiref', host='127.0.0.1', port=8000, **kargs):

    if server in server_names:
        server = server_names.get(server)
    if isinstance(server, type):
        server = server(host=host, port=port, **kargs)
    if not isinstance(server, ServerAdapter):
        raise ValueError("Unknown or unsupported server: %r" % server)

    server.run(app)



import os
from whitenoise import WhiteNoise
from whitenoise.utils import decode_path_info

from .exceptions import HTTPNotFound
from .log import log


PACKAGE_STATICS = os.path.join(os.path.dirname(__file__), "static")


class StaticHandler:
    """ Static handler """

    def __init__(self, root_dir=PACKAGE_STATICS, prefix='static'):
        """ Init"""

        self.whitenoise = WhiteNoise(application=None, root=root_dir, prefix=prefix)

    def add_files(self, path, prefix='static'):
        """ Add files"""

        self.whitenoise.add_files(path, prefix)

    def serve_static(self, environ, start_response):

        path = decode_path_info(environ['PATH_INFO'])
        if self.whitenoise.autorefresh:
            static_file = self.whitenoise.find_file(path)
        else:
            static_file = self.whitenoise.files.get(path)
        if static_file is None:
            raise HTTPNotFound()
        else:
            return self.whitenoise.serve(static_file, environ, start_response)



from collections import namedtuple

BluprintRoute = namedtuple('Route',
                           ['path', 'name', 'handler', 'methods', 'context', 'on_request', 'on_response'])

BluprintStatic = namedtuple('Route',
                            ['route', 'path'])


class Blueprint:
    def __init__(self, name, url_prefix=None, host=None):
        """ """
        self.name = name
        self.url_prefix = url_prefix
        self.host = host

        self.routes = []
        self.statics = []

    def route(self, path,
              name=None,
              methods=frozenset({"GET"}),
              context=None,
              on_request=None,
              on_response=None):
        """ route """

        def decorator(handler):
            self.add_url_rule(path, name, handler, methods, context, on_request, on_response)
            return handler
        return decorator

    def add_url_rule(self, path,
                     name=None,
                     handler=None,
                     methods=frozenset({"GET"}),
                     context=None,
                     on_request=None,
                     on_response=None):
        """ """

        route = BluprintRoute(path, name, handler, methods, context, on_request, on_response)
        self.routes.append(route)

    def static(self, route, path):
        """ get static"""

        route = BluprintStatic(route, path)
        self.statics.append(route)

    def register(self, app, url_prefix=''):
        """ register on the app """

        # Routes
        for route in self.routes:
            # attach the blueprint name to the handler so that it can be
            # prefixed properly in the router
            route.handler.__blueprintname__ = self.name
            # Prepend the blueprint URI prefix if available
            if url_prefix:
                r_path = url_prefix + route.path
            else:
                r_path = route.path
            app.add_url_rule(r_path,
                             name=route.name,
                             handler=route.handler,
                             methods=route.methods,
                             context=route.context,
                             on_request=route.on_request,
                             on_response=route.on_response)

        # statics
        for static in self.statics:
            if url_prefix:
                s_route = url_prefix + static.route
            else:
                s_route = static.route

            app.static(s_route, static.path)



import logging
import logging.config
from inspect import stack, getmodulename
from traceback import format_exc

from kaira.log import log
from kaira.config import Config, LOGGING
from kaira.router import RouteHandler, PathRouter, RouteMethodHandler
from kaira.server import run
from kaira.wrapper import WrapWithContextManager
from kaira.request import HTTPRequest, HTTP_METHODS
from kaira.exceptions import BaseExceptionHTTP, ERR_DETAIL, ERR_CSS, ERR_TEMPLATE, \
    HTTPNotFound, HTTPMethodNotAllowed, HTTPError
from kaira.response import response as kaira_response
from kaira.statics import StaticHandler


class App:

    error_routes = dict()
    static_routes = dict()

    def __init__(self,
                 name=None,
                 router=None,
                 load_env=True,
                 log_config=LOGGING):
        if log_config:
            logging.config.dictConfig(log_config)
        # Only set up a default log handler if the
        # end-user application didn't set anything up.
        if not (logging.root.handlers and
                log.level == logging.NOTSET and
                log_config):
            formatter = logging.Formatter(
                "%(asctime)s: %(levelname)s: %(message)s")
            handler = logging.StreamHandler()
            handler.setFormatter(formatter)
            log.addHandler(handler)
            log.setLevel(logging.INFO)

        # Get name from previous stack frame
        if name is None:
            frame_records = stack()[1]
            name = getmodulename(frame_records[1])

        self.name = name
        self.config = Config(load_env=load_env)
        self.router = router or PathRouter()
        self.debug = None
        self.static_handler = None

    def static(self, route, path):
        """ get static"""

        if not self.static_handler:  # first time
            self.static_handler = StaticHandler(path, prefix=route)
            self.static_routes[route] = path
        else:
            if route not in self.static_routes:
                self.static_handler.add_files(path, prefix=route)
                self.static_routes[route] = path

    def error(self, status_code):
        """
            @app.error(500)
            def e_500():
                return 'Hello World'        
        """

        if 400 <= status_code <= 505:
            pass
        else:
            raise ValueError("Status code must be 400 <= code <= 505")

        def decorator(f):
            self.error_routes[status_code] = f
            return f

        return decorator

    def path_for(self, name, **kwargs):
        """ path for """

        return self.router.path_for(name, **kwargs)

    def route(self, path,
              name=None,
              methods=frozenset({"GET"}),
              context=None,
              on_request=None,
              on_response=None):
        """
            @app.route('/')
            def index():
                return 'Hello World'        
        """

        def decorator(f):
            self.add_url_rule(path,
                              name=name,
                              handler=f,
                              methods=methods,
                              context=context,
                              on_request=on_request,
                              on_response=on_response)
            return f

        return decorator

    def add_url_rule(self, path,
                     name=None,
                     handler=None,
                     methods=frozenset({"GET"}),
                     context=None,
                     on_request=None,
                     on_response=None):
        """ """

        if not path.endswith('/'):
            path += '/'

        methods = frozenset(map(lambda x: x.upper(), methods))

        if not frozenset(methods).issubset(frozenset(HTTP_METHODS)):
            raise ValueError("Not valid method. Only ('GET', 'POST', 'PUT', "
                             "'HEAD', 'OPTIONS', 'PATCH', 'DELETE')")

        wrapped_func = WrapWithContextManager(context)(handler)

        # if already exist
        match_obj, match_dict = self.router.match(path)
        if match_obj:
            route_hand = match_obj
        else:
            route_hand = RouteHandler(path=path,
                                      name=name,
                                      route_func=handler,
                                      methods=methods
                                      )

        for method in methods:
            route_method = RouteMethodHandler(func=wrapped_func,
                                              method=method,
                                              context=context,
                                              on_request=on_request,
                                              on_response=on_response)
            route_hand.handlers[method] = route_method

        if not match_obj:
            self.router.add_route(path, route_hand, name=route_hand.name)

        log.info('Route added. Path: %s Name: %s' % (path, route_hand.name))

    def dynamic_handler(self, environ, start_response):
        """Dynamic handler"""

        request = HTTPRequest(environ, 'utf-8')
        request.app = self

        try:
            response = self.run_dispatcher(request)
        except HTTPError as e:
            if e.content_type in ['text']:
                response = kaira_response.text(e.detail, e.status_code,
                                               cookies=e.cookies, headers=e.headers)
            elif e.content_type in ['json']:
                response = kaira_response.json(e.detail, e.status_code,
                                               cookies=e.cookies, headers=e.headers)

        except BaseExceptionHTTP as e:
            if 400 <= e.status_code <= 505:
                if e.status_code in self.error_routes:
                    err_func = self.error_routes[e.status_code]
                    response = err_func(request)
                else:
                    err_css = ERR_CSS
                    err_detail = ERR_DETAIL[e.status_code]
                    err_code = e.status_code
                    err_explain = 'Oops error ocurred!.'
                    html_template = ERR_TEMPLATE.format(status_code=err_code,
                                                        status_detail=err_detail,
                                                        status_explain=err_explain,
                                                        err_css=err_css)
                    response = kaira_response.html(html_template, status_code=e.status_code)
            elif 300 <= e.status_code <= 307:
                response = kaira_response.redirect(e.absolute_url, e.status_code,
                                                   cookies=e.cookies, headers=e.headers)
            else:
                raise NotImplemented("Invalid status code only 300 or 400 http codes")

        except Exception as e:
            err_code = 500
            err_css = ERR_CSS
            err_detail = ERR_DETAIL[err_code]
            err_explain_debug = "Error while handling error: {}\nStack: {}".format(
                e, format_exc())
            err_explain = "An error occurred while handling an error"
            if self.debug:
                err_explain = err_explain_debug
            log.error(err_explain_debug)
            html_template = ERR_TEMPLATE.format(status_code=err_code,
                                                status_detail=err_detail,
                                                status_explain=err_explain,
                                                err_css=err_css)
            response = kaira_response.html(html_template, status_code=err_code)

        return response.__call__(start_response)

    def run_dispatcher(self, request):
        """ Run dipatcher """

        response = None
        if not request.path.endswith('/'):
            request.path += '/'
        match_obj, match_dict = self.router.match(request.path)
        if match_obj:

            try:
                del match_dict['route_name']  # remove route_name key from dictionary
            except KeyError:
                pass

            try:
                method_handler = match_obj.handlers[request.method]
            except KeyError:
                raise HTTPMethodNotAllowed()

            if method_handler.on_request:
                for func_on_req in method_handler.on_request:
                    func_on_req(request)

            response = method_handler.func(request, **match_dict)

            if method_handler.on_response:
                for func_on_resp in method_handler.on_response:
                    func_on_resp(request, response)

        else:
            raise HTTPNotFound()

        return response

    def serve_static_handler(self, environ, start_response):
        """static handler"""

        try:
            response = self.static_handler.serve_static(environ, start_response)
            return response

        except BaseExceptionHTTP as e:
            request = HTTPRequest(environ, 'utf-8')
            if 400 <= e.status_code <= 505:
                if e.status_code in self.error_routes:
                    err_func = self.error_routes[e.status_code]
                    response = err_func(request)
                else:
                    err_css = ERR_CSS
                    err_detail = ERR_DETAIL[e.status_code]
                    err_code = e.status_code
                    err_explain = 'Oops error ocurred!.'
                    html_template = ERR_TEMPLATE.format(status_code=err_code,
                                                        status_detail=err_detail,
                                                        status_explain=err_explain,
                                                        err_css=err_css)
                    response = kaira_response.html(html_template, status_code=e.status_code)
            elif 300 <= e.status_code <= 307:
                kaira_response.redirect(e.absolute_url, e.status_code)
            else:
                raise NotImplemented("Invalid status code only 300 or 400 http codes")

            return response.__call__(start_response)

        except Exception as e:
            request = HTTPRequest(environ, 'utf-8')
            err_code = 500
            err_css = ERR_CSS
            err_detail = ERR_DETAIL[err_code]
            err_explain_debug = "Error while handling error: {}\nStack: {}".format(
                e, format_exc())
            err_explain = "An error occurred while handling an error"
            log.error(err_explain)
            if self.debug:
                err_explain = err_explain_debug
            html_template = ERR_TEMPLATE.format(status_code=err_code,
                                                status_detail=err_detail,
                                                status_explain=err_explain,
                                                err_css=err_css)
            response = kaira_response.html(html_template, status_code=err_code)

            return response.__call__(start_response)

    def wsgi_app(self, environ, start_response):
        """ """

        path_info = environ['PATH_INFO']
        registered_routes = self.static_routes.keys()
        static_match = path_info.startswith(tuple(registered_routes))

        if static_match:  # check if visitor has requested a static file
            return self.serve_static_handler(environ, start_response)
        else:  # else call the dynamic handler
            return self.dynamic_handler(environ, start_response)

    def __call__(self, environ, start_response):
        """Shortcut for :attr:`wsgi_app`."""

        return self.wsgi_app(environ, start_response)

    def run(self, debug=True, **kwargs):

        self.debug = debug
        log.info("Server started.. listening on host: %s:%s" % (kwargs['host'], kwargs['port']))
        run(self, **kwargs)

    def __repr__(self):
        return '<%s %r>' % (
            self.__class__.__name__,
            self.name,
        )




import base64
import datetime
import json
import re
import time
import zlib
from decimal import Decimal
import hmac
import hashlib

from kaira.baseconv import base62

_PROTECTED_TYPES = (
    type(None), int, float, Decimal, datetime.datetime, datetime.date, datetime.time,
)
_SEP_UNSAFE = re.compile(r'^[A-z0-9-_=]*$')


def is_protected_type(obj):
    """Determine if the object instance is of a protected type.
    Objects of protected types are preserved as-is when passed to
    force_text(strings_only=True).
    """
    return isinstance(obj, _PROTECTED_TYPES)


def force_bytes(s, encoding='utf-8', strings_only=False, errors='strict'):
    """
    Similar to smart_bytes, except that lazy instances are resolved to
    strings, rather than kept as lazy objects.
    If strings_only is True, don't convert (some) non-string-like objects.
    """
    # Handle the common case first for performance reasons.
    if isinstance(s, bytes):
        if encoding == 'utf-8':
            return s
        else:
            return s.decode('utf-8', errors).encode(encoding, errors)
    if strings_only and is_protected_type(s):
        return s
    if isinstance(s, memoryview):
        return bytes(s)
    if not isinstance(s, str):
        return str(s).encode(encoding, errors)
    else:
        return s.encode(encoding, errors)


class BadSignature(Exception):
    """Signature does not match."""
    pass


class SignatureExpired(BadSignature):
    """Signature timestamp is older than required max_age."""
    pass


def salted_hmac(key_salt, value, secret=None):
    """
    Return the HMAC-SHA1 of 'value', using a key generated from key_salt and a
    secret (which defaults to settings.SECRET_KEY).
    A different key_salt should be passed in for every application of HMAC.
    """
    if secret is None:
        secret = b'APj00jpfj8Gx1SjnyLxwBBSQfnQ9DJYe0Cm'

    key_salt = force_bytes(key_salt)
    secret = force_bytes(secret)

    # We need to generate a derived key from our base key.  We can do this by
    # passing the key_salt and our base key through a pseudo-random function and
    # SHA1 works nicely.
    key = hashlib.sha1(key_salt + secret).digest()

    # If len(key_salt + secret) > sha_constructor().block_size, the above
    # line is redundant and could be replaced by key = key_salt + secret, since
    # the hmac module does the same thing for keys longer than the block size.
    # However, we need to ensure that we *always* do this.
    return hmac.new(key, msg=force_bytes(value), digestmod=hashlib.sha1)


def constant_time_compare(val1, val2):
    """Return True if the two strings are equal, False otherwise."""
    return hmac.compare_digest(force_bytes(val1), force_bytes(val2))


def b64_encode(s):
    return base64.urlsafe_b64encode(s).strip(b'=')


def b64_decode(s):
    pad = b'=' * (-len(s) % 4)
    return base64.urlsafe_b64decode(s + pad)


def base64_hmac(salt, value, key):
    return b64_encode(salted_hmac(salt, value, key).digest()).decode()


class JSONSerializer:
    """
    Simple wrapper around json to be used in signing.dumps and
    signing.loads.
    """
    def dumps(self, obj):
        return json.dumps(obj, separators=(',', ':')).encode('latin-1')

    def loads(self, data):
        return json.loads(data.decode('latin-1'))


def dumps(obj, key=None, salt='kaira.sessions', serializer=JSONSerializer, compress=False):
    """
    Return URL-safe, hmac/SHA1 signed base64 compressed JSON string. If key is
    None, use settings.SECRET_KEY instead.
    If compress is True (not the default), check if compressing using zlib can
    save some space. Prepend a '.' to signify compression. This is included
    in the signature, to protect against zip bombs.
    Salt can be used to namespace the hash, so that a signed string is
    only valid for a given namespace. Leaving this at the default
    value or re-using a salt value across different parts of your
    application without good cause is a security risk.
    The serializer is expected to return a bytestring.
    """
    data = serializer().dumps(obj)

    # Flag for if it's been compressed or not
    is_compressed = False

    if compress:
        # Avoid zlib dependency unless compress is being used
        compressed = zlib.compress(data)
        if len(compressed) < (len(data) - 1):
            data = compressed
            is_compressed = True
    base64d = b64_encode(data).decode()
    if is_compressed:
        base64d = '.' + base64d
    return TimestampSigner(key, salt=salt).sign(base64d)


def loads(s, key=None, salt='kaira.sessions', serializer=JSONSerializer, max_age=None):
    """
    Reverse of dumps(), raise BadSignature if signature fails.
    The serializer is expected to accept a bytestring.
    """
    # TimestampSigner.unsign() returns str but base64 and zlib compression
    # operate on bytes.
    base64d = TimestampSigner(key, salt=salt).unsign(s, max_age=max_age).encode()
    decompress = base64d[:1] == b'.'
    if decompress:
        # It's compressed; uncompress it first
        base64d = base64d[1:]
    data = b64_decode(base64d)
    if decompress:
        data = zlib.decompress(data)
    return serializer().loads(data)


class Signer:

    def __init__(self, key=None, sep=':', salt=None):
        # Use of native strings in all versions of Python
        self.key = key or 'APj00jpfj8Gx1SjnyLxwBBSQfnQ9DJYe0Cm'
        self.sep = sep
        if _SEP_UNSAFE.match(self.sep):
            raise ValueError(
                'Unsafe Signer separator: %r (cannot be empty or consist of '
                'only A-z0-9-_=)' % sep,
            )
        self.salt = salt or '%s.%s' % (self.__class__.__module__, self.__class__.__name__)

    def signature(self, value):
        return base64_hmac(self.salt + 'signer', value, self.key)

    def sign(self, value):
        return '%s%s%s' % (value, self.sep, self.signature(value))

    def unsign(self, signed_value):
        if self.sep not in signed_value:
            raise BadSignature('No "%s" found in value' % self.sep)
        value, sig = signed_value.rsplit(self.sep, 1)
        if constant_time_compare(sig, self.signature(value)):
            return value
        raise BadSignature('Signature "%s" does not match' % sig)


class TimestampSigner(Signer):

    def timestamp(self):
        return base62.encode(int(time.time()))

    def sign(self, value):
        value = '%s%s%s' % (value, self.sep, self.timestamp())
        return super().sign(value)

    def unsign(self, value, max_age=None):
        """
        Retrieve original value and check it wasn't signed more
        than max_age seconds ago.
        """
        result = super().unsign(value)
        value, timestamp = result.rsplit(self.sep, 1)
        timestamp = base62.decode(timestamp)
        if max_age is not None:
            if isinstance(max_age, datetime.timedelta):
                max_age = max_age.total_seconds()
            # Check timestamp is not older than max_age
            age = time.time() - timestamp
            if age > max_age:
                raise SignatureExpired(
                    'Signature age %s > %s seconds' % (age, max_age))
        return value



import os
import functools


TEMPLATE_PATH = ['./', './views/']
TEMPLATES = {}
DEBUG = False


class TemplateError(Exception):
    pass


class BaseTemplate(object):
    """ Base class and minimal API for template adapters """
    extensions = ['tpl', 'html', 'thtml', 'stpl']
    settings = {}  #used in prepare()
    defaults = {}  #used in render()

    def __init__(self,
                 source=None,
                 name=None,
                 lookup=None,
                 encoding='utf8', **settings):
        """ Create a new template.
        If the source parameter (str or buffer) is missing, the name argument
        is used to guess a template filename. Subclasses can assume that
        self.source and/or self.filename are set. Both are strings.
        The lookup, encoding and settings parameters are stored as instance
        variables.
        The lookup parameter stores a list containing directory paths.
        The encoding parameter should be used to decode byte strings or files.
        The settings parameter contains a dict for engine-specific settings.
        """
        self.name = name
        self.source = source.read() if hasattr(source, 'read') else source
        self.filename = source.filename if hasattr(source, 'filename') else None
        self.lookup = [os.path.abspath(x) for x in lookup] if lookup else []
        self.encoding = encoding
        self.settings = self.settings.copy()  # Copy from class variable
        self.settings.update(settings)  # Apply
        if not self.source and self.name:
            self.filename = self.search(self.name, self.lookup)
            if not self.filename:
                raise TemplateError('Template %s not found.' % repr(name))
        if not self.source and not self.filename:
            raise TemplateError('No template specified.')
        self.prepare(**self.settings)

    @classmethod
    def search(cls, name, lookup=None):
        """ Search name in all directories specified in lookup.
        First without, then with common extensions. Return first hit. """

        for spath in lookup:
            spath = os.path.abspath(spath) + os.sep
            fname = os.path.abspath(os.path.join(spath, name))
            if not fname.startswith(spath): continue
            if os.path.isfile(fname): return fname
            for ext in cls.extensions:
                if os.path.isfile('%s.%s' % (fname, ext)):
                    return '%s.%s' % (fname, ext)

    @classmethod
    def global_config(cls, key, *args):
        """ This reads or sets the global settings stored in class.settings. """
        if args:
            cls.settings = cls.settings.copy()  # Make settings local to class
            cls.settings[key] = args[0]
        else:
            return cls.settings[key]

    def prepare(self, **options):
        """ Run preparations (parsing, caching, ...).
        It should be possible to call this again to refresh a template or to
        update settings.
        """
        raise NotImplementedError

    def render(self, *args, **kwargs):
        """ Render the template with the specified local variables and return
        a single byte or unicode string. If it is a byte string, the encoding
        must match self.encoding. This method must be thread-safe!
        Local variables may be provided in dictionaries (args)
        or directly, as keywords (kwargs).
        """
        raise NotImplementedError


class MakoTemplate(BaseTemplate):
    def prepare(self, **options):
        from mako.template import Template
        from mako.lookup import TemplateLookup
        options.update({'input_encoding': self.encoding})
        options.setdefault('format_exceptions', bool(DEBUG))
        lookup = TemplateLookup(directories=self.lookup, **options)
        if self.source:
            self.tpl = Template(self.source, lookup=lookup, **options)
        else:
            self.tpl = Template(uri=self.name,
                                filename=self.filename,
                                lookup=lookup, **options)

    def render(self, *args, **kwargs):
        for dictarg in args:
            kwargs.update(dictarg)
        _defaults = self.defaults.copy()
        _defaults.update(kwargs)
        return self.tpl.render(**_defaults)


def template(*args, **kwargs):
    """
    Get a rendered template as a string iterator.
    You can use a name, a filename or a template string as first parameter.
    Template rendering arguments can be passed as dictionaries
    or directly (as keyword arguments).
    """
    tpl = args[0] if args else None
    for dictarg in args[1:]:
        kwargs.update(dictarg)
    adapter = kwargs.pop('template_adapter', MakoTemplate)
    lookup = kwargs.pop('template_lookup', TEMPLATE_PATH)
    tplid = (id(lookup), tpl)
    if tplid not in TEMPLATES or DEBUG:
        settings = kwargs.pop('template_settings', {})
        if isinstance(tpl, adapter):
            TEMPLATES[tplid] = tpl
            if settings: TEMPLATES[tplid].prepare(**settings)
        elif "\n" in tpl or "{" in tpl or "%" in tpl or '$' in tpl:
            TEMPLATES[tplid] = adapter(source=tpl, lookup=lookup, **settings)
        else:
            TEMPLATES[tplid] = adapter(name=tpl, lookup=lookup, **settings)
    if not TEMPLATES[tplid]:
        abort(500, 'Template (%s) not found' % tpl)
    return TEMPLATES[tplid].render(kwargs)


#mako_template = functools.partial(template, template_adapter=MakoTemplate)



""" Request module"""

import sys
from cgi import FieldStorage
import json
from urllib.parse import urlunsplit
from urllib.parse import unquote

try:
    from ujson import loads as json_loads
except ImportError:
    if sys.version_info[:2] == (3, 5):
        def json_loads(data):
            # on Python 3.5 json.loads only supports str not bytes
            return json.loads(data.decode())
    else:
        json_loads = json.loads

try:  # pragma: nocover
    # Python 2.5+
    partition = str.partition
except AttributeError:  # pragma: nocover
    def partition(s, sep):
        if sep in s:
            a, b = s.split(sep, 1)
            return a, sep, b
        else:
            return s, sep, ''


MULTIPART_ENVIRON = {'REQUEST_METHOD': 'POST'}
HTTP_METHODS = ('GET', 'POST', 'PUT', 'HEAD', 'OPTIONS', 'PATCH', 'DELETE')


class attribute(object):
    """ ``attribute`` decorator is intended to promote a
        function call to object attribute. This means the
        function is called once and replaced with
        returned value.

        >>> class A:
        ...     def __init__(self):
        ...         self.counter = 0
        ...     @attribute
        ...     def count(self):
        ...         self.counter += 1
        ...         return self.counter
        >>> a = A()
        >>> a.count
        1
        >>> a.count
        1
    """
    __slots__ = ('f')

    def __init__(self, f):
        self.f = f

    def __get__(self, obj, t=None):
        f = self.f
        val = f(obj)
        setattr(obj, f.__name__, val)
        return val


class HTTPRequest(object):
    """  """

    app = None
    auth = None

    def __init__(self, environ: dict, encoding: str, max_content_lenght: int=4*1024*1024*1024) -> None:
        self.environ = environ
        self.encoding = encoding
        self.max_content_lenght = max_content_lenght
        self.method = environ['REQUEST_METHOD']

    @attribute
    def host(self) -> str:
        host = self.environ['HTTP_HOST']
        if ',' in host:
            host = host.rsplit(',', 1)[-1].strip()
        return host

    @attribute
    def remote_addr(self) -> str:
        addr = self.environ['REMOTE_ADDR']
        if ',' in addr:
            addr = addr.split(',', 1)[0].strip()
        return addr

    @attribute
    def root_path(self) -> str:
        return self.environ['SCRIPT_NAME'] + '/'

    @attribute
    def path(self) -> str:
        return self.environ['SCRIPT_NAME'] + self.environ['PATH_INFO']

    @attribute
    def query(self) -> str:
        return parse_qs(self.environ['QUERY_STRING'])

    def get_param(self, name):
        p = self.query.get(name)
        return p and p[-1]

    @attribute
    def form(self):
        form, self.files = self.load_body()
        return form

    @attribute
    def files(self):
        self.form, files = self.load_body()
        return files

    @attribute
    def json(self):
        json_b, self.files = self.load_body()
        return json_b

    @attribute
    def cookies(self):
        if 'HTTP_COOKIE' in self.environ:
            return parse_cookie(self.environ['HTTP_COOKIE'])
        else:
            return {}

    @attribute
    def ajax(self):
        if 'HTTP_X_REQUESTED_WITH' in self.environ:
            return self.environ['HTTP_X_REQUESTED_WITH'] == 'XMLHttpRequest'
        else:
            return False

    @attribute
    def secure(self) -> bool:
        return self.environ['wsgi.url_scheme'] == 'https'

    @attribute
    def scheme(self) -> str:
        return self.environ['wsgi.url_scheme']

    @attribute
    def urlparts(self):
        return UrlParts((self.scheme, self.host,
                         self.path, self.environ['QUERY_STRING'], None))

    @attribute
    def content_type(self) -> str:
        return self.environ['CONTENT_TYPE']

    @attribute
    def content_length(self) -> int:
        return int(self.environ['CONTENT_LENGTH'])

    @attribute
    def stream(self):
        return self.environ['wsgi.input']

    def load_body(self):
        """ Load http request body and returns
            form data and files.
        """
        environ = self.environ
        cl = environ['CONTENT_LENGTH']
        icl = int(cl)
        if icl > self.max_content_lenght:
            raise ValueError('Maximum content length exceeded')
        fp = environ['wsgi.input']
        ct = environ['CONTENT_TYPE']
        # application/x-www-form-urlencoded
        if '/x' in ct:
            return parse_qs(fp.read(icl).decode(self.encoding)), None
        # application/json
        elif '/j' in ct:
            return json_loads(fp.read(icl).decode(self.encoding)), None
        # multipart/form-data
        elif ct.startswith('m'):
            return parse_multipart(fp, ct, cl, self.encoding)
        else:
            return None, None


def parse_qs(qs):
    params = {}
    for field in qs.split('&'):
        r = partition(field, '=')
        k = r[0]
        v = r[2]
        if '+' in k:
            k = k.replace('+', ' ')
        if '%' in k:
            k = unquote(k)
        if '+' in v:
            v = v.replace('+', ' ')
        if k in params:
            params[k].append('%' in v and unquote(v) or v)
        else:
            if ',' in v:
                params[k] = [('%' in v and unquote(x) or x)
                             for x in v.split(',')]
            else:
                params[k] = ['%' in v and unquote(v) or v]
    return params


def parse_multipart(fp, ctype, clength, encoding):
    """ Parse multipart/form-data request. Returns
        a tuple (form, files).
    """
    fs = FieldStorage(
        fp=fp,
        environ=MULTIPART_ENVIRON,
        headers={
            'content-type': ctype,
            'content-length': clength
        },
        keep_blank_values=True
    )
    form = {}
    files = {}
    for f in fs.list:
        if f.filename:
            files.setdefault(f.name, []).append(f)
        else:
            form.setdefault(f.name, []).append(f.value)
    return form, files


def parse_cookie(cookie):
    """ Parse cookie string and return a dictionary
        where key is a name of the cookie and value
        is cookie value.
    """
    return cookie and dict([pair.split('=', 1)
                            for pair in cookie.split('; ')]) or {}


def urlparts(parts=None, scheme=None, netloc=None, path=None,
             query=None, fragment=None):
    """  """
    if not parts:
        parts = (scheme, netloc, path, query, fragment)
    return UrlParts(parts)


class UrlParts(tuple):
    """ Concrete class for :func:`urlparse.urlsplit` results.
    """

    def __init__(self, parts):
        assert len(parts) == 5, '`parts` must be a tupple of length 6'
        super(UrlParts, self).__init__()

    def __repr__(self):
        return 'urlparts' + super(UrlParts, self).__repr__()

    def geturl(self):
        """ """
        return urlunsplit(self)

    def join(self, other):
        """  """
        parts = (
            other[0] or self[0],
            other[1] or self[1],
            other[2] or self[2],
            other[3],
            other[4])
        return UrlParts(parts)




""" ``cookie`` module.
"""

from datetime import datetime
from time import time

from kaira.utils import format_http_datetime, n


class HTTPCookie(object):
    """ HTTP Cookie
        http://www.ietf.org/rfc/rfc2109.txt

        ``domain``, ``secure`` and ``httponly`` are
        taken from ``config`` if not set.
    """
    __slots__ = ('name', 'value', 'path', 'expires',
                 'domain', 'secure', 'httponly')

    def __init__(self, name, value=None, path='/',
                 expires=None, max_age=None,
                 domain=None, secure=None, httponly=None,
                 options=None):
        self.name = name
        self.value = value
        self.path = path
        if max_age is None:
            self.expires = expires
        else:
            self.expires = datetime.utcfromtimestamp(time() + max_age)
        if domain is None:
            self.domain = options['HTTP_COOKIE_DOMAIN']
        else:
            self.domain = domain
        if secure is None:
            self.secure = options['HTTP_COOKIE_SECURE']
        else:
            self.secure = secure
        if httponly is None:
            self.httponly = options['HTTP_COOKIE_HTTPONLY']
        else:
            self.httponly = httponly

    @classmethod
    def delete(cls, name, path='/', domain=None, options=None):
        """ Returns a cookie to be deleted by browser.
        """
        return cls(name,
                   expires='Sat, 01 Jan 2000 00:00:01 GMT',
                   path=path, domain=domain, options=options)

    def http_set_cookie(self, encoding):
        """ Returns Set-Cookie response header.
        """
        directives = []
        append = directives.append
        append(self.name + '=')
        if self.value:
            append(n(self.value, encoding))
        if self.domain:
            append('; domain=' + self.domain)
        if self.expires:
            append('; expires=' + format_http_datetime(self.expires))
        if self.path:
            append('; path=' + self.path)
        if self.secure:
            append('; secure')
        if self.httponly:
            append('; httponly')
        return ('Set-Cookie', ''.join(directives))


class CookieManager(dict):
    """  """

    def __init__(self, options):
        super().__init__()
        self.options = options

    def __setitem__(self, key, value):
        # If this cookie doesn't exist, add it to the header keys
        cookie = HTTPCookie(key, value, options=self.options)
        return super().__setitem__(key, cookie)

    def __delitem__(self, key):

        if key not in self:
            cookie = HTTPCookie(key, '', options=self.options)
            cookie.expires = 'Sat, 01 Jan 2000 00:00:01 GMT'
            #self[key] = cookie
            super().__setitem__(key, cookie)
        else:
            #self[key] = ''
            self[key].expires = 'Sat, 01 Jan 2000 00:00:01 GMT'




from time import localtime
from time import mktime
import html
import urllib.parse

from datetime import datetime
from datetime import timedelta
from datetime import tzinfo


ZERO = timedelta(0)
WEEKDAYS = ("Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun")
MONTHS = (
    None, "Jan", "Feb", "Mar", "Apr", "May", "Jun",
    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec")


bytes_type = bytes
str_type = str


def sanitize(text):
    """Gets rid of < and > and & and, for good measure, :"""

    return html.escape(text, quote=True).replace(':', '&#58;')


def n(s, encoding='latin1'):
    if isinstance(s, str_type):
        return s
    return s.decode(encoding)


def format_http_datetime(stamp):
    """ Formats datetime to a string following rfc1123 pattern.

        >>> now = datetime(2011, 9, 19, 10, 45, 30, 0, UTC)
        >>> format_http_datetime(now)
        'Mon, 19 Sep 2011 10:45:30 GMT'

        if timezone is not set in datetime instance the ``stamp``
        is assumed to be in UTC (``datetime.utcnow``).

        >>> now = datetime(2011, 9, 19, 10, 45, 30, 0)
        >>> format_http_datetime(now)
        'Mon, 19 Sep 2011 10:45:30 GMT'

        >>> now = datetime.utcnow()
        >>> assert format_http_datetime(now)

        if ``stamp`` is a string just return it

        >>> format_http_datetime('x')
        'x'

        >>> format_http_datetime(100) # doctest: +ELLIPSIS
        Traceback (most recent call last):
            ...
        TypeError: ...
    """
    if isinstance(stamp, datetime):
        if stamp.tzinfo:
            stamp = stamp.astimezone(UTC).timetuple()
        else:
            stamp = localtime(mktime(stamp.timetuple()))
    elif isinstance(stamp, str):
        return stamp
    else:
        raise TypeError('Expecting type ``datetime.datetime``.')

    year, month, day, hh, mm, ss, wd, y, z = stamp
    return "%s, %02d %3s %4d %02d:%02d:%02d GMT" % (
        WEEKDAYS[wd], day, MONTHS[month], year, hh, mm, ss
    )


class utc(tzinfo):
    """ UTC timezone.

    """
    __slots__ = ('name')

    def __init__(self, name):
        self.name = name

    def tzname(self, dt):
        """ Name of time zone.

            >>> GMT.tzname(None)
            'GMT'
            >>> UTC.tzname(None)
            'UTC'
        """
        return self.name

    def utcoffset(self, dt):
        """ Offset from UTC.

            >>> UTC.utcoffset(None)
            datetime.timedelta(0)
        """
        return ZERO

    def dst(self, dt):
        """ DST is not in effect.

            >>> UTC.dst(None)
            datetime.timedelta(0)
        """
        return ZERO


GMT = utc('GMT')
UTC = utc('UTC')


def url(*args, **kwargs):
    """Url make"""

    func_par = args
    if 'args' in kwargs:
        func_args = kwargs['args']
    else:
        func_args = None

    if 'vars' in kwargs:
        func_vars = kwargs['vars']
    else:
        func_vars = None

    if func_args in (None, []):
        func_args = []

    if not func_par:
        raise SyntaxError("You need to pass at least one argument")

    func_vars = func_vars or {}

    if not isinstance(func_par, (list, tuple)):
        func_par = [func_par]

    if not isinstance(func_args, (list, tuple)):
        func_args = [func_args]

    if func_args:
        other = func_args and urllib.parse.quote(
                '/' + '/'.join([str(x) for x in func_args]))
    else:
        other = ''

    if other.endswith('/'):
        other += '/'    # add trailing slash to make last trailing empty arg explicit

    list_vars = []
    for (key, vals) in sorted(func_vars.items()):
        if not isinstance(vals, (list, tuple)):
            vals = [vals]
        for val in vals:
            list_vars.append((key, val))

    if list_vars:
        other += '?%s' % urllib.parse.urlencode(list_vars)

    func_and_par = func_par and urllib.parse.quote(
        '/' + '/'.join([str(x) for x in func_par]))

    func_and_par = func_and_par.replace('//', '/')

    return func_and_par + other




from distutils.core import setup


def readme():
    with open('README.rst') as f:
        return f.read()


setup(
    name='kaira',
    version='0.2.2',
    packages=['kaira', 'kaira.routing'],
    url='https://github.com/mulonemartin/kaira/',
    license='MIT',
    author='Martin Mulone',
    author_email='mulone.martin@gmail.com',
    description='Web Micro framework - Python',
    long_description=readme(),
    install_requires=[
              'whitenoise', 'wtforms', 'multidict'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Topic :: Internet :: WWW/HTTP',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6'
    ],
    include_package_data=True,
    zip_safe=False
)




class BaseExceptionHTTP(Exception):

    default_status_code = 500
    default_detail = 'Server error'

    def __init__(self, detail=None, status_code=None):
        super().__init__(detail)

        self.detail = self.default_detail if (detail is None) else detail
        self.status_code = self.default_status_code if (status_code is None) else status_code


class HTTPError(BaseExceptionHTTP):
    default_status_code = 500
    default_detail = 'Server error'
    cookies = None
    headers = None
    content_type = 'text'

    def __init__(self, detail=None, status_code=None, cookies=None, headers=None, content_type='text'):
        super().__init__(detail=detail, status_code=status_code)

        self.cookies = cookies
        self.header = headers
        self.content_type = content_type


class HTTPRedirect(BaseExceptionHTTP):

    default_status_code = 302
    default_detail = ''
    cookies = None
    headers = None

    def __init__(self, absolute_url, detail=None, status_code=None, cookies=None, headers=None):

        super().__init__(detail, status_code)
        self.absolute_url = absolute_url
        self.cookies = cookies
        self.header = headers


class HTTPNotFound(BaseExceptionHTTP):
    default_status_code = 404
    default_detail = 'Not found'


class HTTPMethodNotAllowed(BaseExceptionHTTP):
    default_status_code = 405
    default_detail = 'Method Not Allowed'


class HandleException:

    @staticmethod
    def redirect(absolute_url, detail=None, status_code=None, cookies=None, headers=None):
        """ redirect """

        resp = HTTPRedirect(absolute_url, detail=detail, status_code=status_code,
                            cookies=cookies, headers=headers)
        return resp

    @staticmethod
    def json(body, status_code=None, cookies=None, headers=None):
        """ json """

        resp = HTTPError(body, status_code=status_code, cookies=cookies,
                         headers=headers, content_type='json')
        return resp

    @staticmethod
    def text(body, status_code=None, cookies=None, headers=None):
        """ text """

        resp = HTTPError(body, status_code=status_code, cookies=cookies,
                         headers=headers, content_type='text')
        return resp


http_exception = HandleException()


ERR_DETAIL = {
    # Client Error
    400: 'Bad Request',
    401: 'Unauthorized',
    402: 'Payment Required',
    403: 'Forbidden',
    404: 'Not Found',
    405: 'Method Not Allowed',
    406: 'Not Acceptable',
    407: 'Proxy Authentication Required',
    408: 'Request Timeout',
    409: 'Conflict',
    410: 'Gone',
    411: 'Length Required',
    412: 'Precondition Failed',
    413: 'Request Entity Too Large',
    414: 'Request-Uri Too Long',
    415: 'Unsupported Media Type',
    416: 'Requested Range Not Satisfiable',
    417: 'Expectation Failed',
    # Server Error
    500: 'Internal Server Error',
    501: 'Not Implemented',
    502: 'Bad Gateway',
    503: 'Service Unavailable',
    504: 'Gateway Timeout',
    505: 'Http Version Not Supported'
}

ERR_CSS = '''
<style type="text/css">/*! normalize.css v5.0.0 | MIT License | github.com/necolas/normalize.css */html{font-family:sans-serif;line-height:1.15;-ms-text-size-adjust:100%;-webkit-text-size-adjust:100%}body{margin:0}article,aside,footer,header,nav,section{display:block}h1{font-size:2em;margin:.67em 0}figcaption,figure,main{display:block}figure{margin:1em 40px}hr{box-sizing:content-box;height:0;overflow:visible}pre{font-family:monospace,monospace;font-size:1em}a{background-color:transparent;-webkit-text-decoration-skip:objects}a:active,a:hover{outline-width:0}abbr[title]{border-bottom:none;text-decoration:underline;text-decoration:underline dotted}b,strong{font-weight:inherit}b,strong{font-weight:bolder}code,kbd,samp{font-family:monospace,monospace;font-size:1em}dfn{font-style:italic}mark{background-color:#ff0;color:#000}small{font-size:80%}sub,sup{font-size:75%;line-height:0;position:relative;vertical-align:baseline}sub{bottom:-.25em}sup{top:-.5em}audio,video{display:inline-block}audio:not([controls]){display:none;height:0}img{border-style:none}svg:not(:root){overflow:hidden}button,input,optgroup,select,textarea{font-family:sans-serif;font-size:100%;line-height:1.15;margin:0}button,input{overflow:visible}button,select{text-transform:none}[type=reset],[type=submit],button,html [type=button]{-webkit-appearance:button}[type=button]::-moz-focus-inner,[type=reset]::-moz-focus-inner,[type=submit]::-moz-focus-inner,button::-moz-focus-inner{border-style:none;padding:0}[type=button]:-moz-focusring,[type=reset]:-moz-focusring,[type=submit]:-moz-focusring,button:-moz-focusring{outline:1px dotted ButtonText}fieldset{border:1px solid silver;margin:0 2px;padding:.35em .625em .75em}legend{box-sizing:border-box;color:inherit;display:table;max-width:100%;padding:0;white-space:normal}progress{display:inline-block;vertical-align:baseline}textarea{overflow:auto}[type=checkbox],[type=radio]{box-sizing:border-box;padding:0}[type=number]::-webkit-inner-spin-button,[type=number]::-webkit-outer-spin-button{height:auto}[type=search]{-webkit-appearance:textfield;outline-offset:-2px}[type=search]::-webkit-search-cancel-button,[type=search]::-webkit-search-decoration{-webkit-appearance:none}::-webkit-file-upload-button{-webkit-appearance:button;font:inherit}details,menu{display:block}summary{display:list-item}canvas{display:inline-block}template{display:none}[hidden]{display:none}/*! Simple HttpErrorPages | MIT X11 License | https://github.com/AndiDittrich/HttpErrorPages */body,html{width:100%;height:100%;background-color:#21232a}body{color:#fff;text-align:center;text-shadow:0 2px 4px rgba(0,0,0,.5);padding:0;min-height:100%;-webkit-box-shadow:inset 0 0 75pt rgba(0,0,0,.8);box-shadow:inset 0 0 75pt rgba(0,0,0,.8);display:table;font-family:"Open Sans",Arial,sans-serif}h1{font-family:inherit;font-weight:500;line-height:1.1;color:inherit;font-size:36px}h1 small{font-size:68%;font-weight:400;line-height:1;color:#777}a{text-decoration:none;color:#fff;font-size:inherit;border-bottom:dotted 1px #707070}.lead{color:silver;font-size:21px;line-height:1.4}.cover{display:table-cell;vertical-align:middle;padding:0 20px}footer{position:fixed;width:100%;height:40px;left:0;bottom:0;color:#a0a0a0;font-size:14px}</style>
'''

ERR_TEMPLATE = '''
<!DOCTYPE html>
<html lang="en">
<head>
    <!-- Simple HttpErrorPages | MIT X11 License | https://github.com/AndiDittrich/HttpErrorPages -->

    <meta charset="utf-8" />
    <meta http-equiv="X-UA-Compatible" content="IE=edge" />
    <meta name="viewport" content="width=device-width, initial-scale=1" />
    
    <title>We've got some trouble | {status_code} - {status_detail}</title>

    {err_css}
    
</head>

<body>
    <div class="cover">
        <h1>{status_detail} <small>Error {status_code}</small></h1>
        <p class="lead">{status_explain}.</p>
    </div>
    
    </body>
</html>

'''



""" Messages module """

from datetime import timedelta
import datetime

# Kaira import
from kaira.signing import dumps as signing_dumps
from kaira.signing import loads as signing_loads
from kaira.cookie import CookieManager
from kaira.wrapper import ContextManager


DEBUG = 10
INFO = 20
SUCCESS = 25
WARNING = 30
ERROR = 40

DEFAULT_TAGS = {
    DEBUG: 'debug',
    INFO: 'info',
    SUCCESS: 'success',
    WARNING: 'warning',
    ERROR: 'error',
}

DEFAULT_LEVELS = {
    'DEBUG': DEBUG,
    'INFO': INFO,
    'SUCCESS': SUCCESS,
    'WARNING': WARNING,
    'ERROR': ERROR,
}
"""
messages = [{'msg': 'hola mundo', 'lvl': 20}, {'msg': 'hola mundo 2', 'lvl': 20}]
"""


class MessagesManager:

    # uwsgi's default configuration enforces a maximum size of 4kb for all the
    # HTTP headers. In order to leave some room for other cookies and headers,
    # restrict the session cookie to 1/2 of 4kb. See #18781.
    max_cookie_size = 2048
    request = None
    cookies = None
    _queued_messages = []
    used = False
    messages = []
    response_cookies = None

    def __init__(self, options=None):

        if not options:
            options = {
                       'MESSAGES_COOKIE_DOMAIN': '',
                       'MESSAGES_COOKIE_SECURE': False,
                       'MESSAGES_COOKIE_HTTPONLY': True,
                       'MESSAGES_COOKIE_EXPIRE': 600,
                       'MESSAGES_COOKIE_PATH': '/',
                       'MESSAGES_COOKIE_NAME': 'kaira_message',
                       'MESSAGES_SECRET': b'APj00cpfa8Gx1SjnyLxwBBSQfnQ9DJYe0Cm',
                       'MESSAGES_TIME_LIMIT': timedelta(minutes=60)
                      }
        self.options = options

    def start(self, request, cookies=None):
        self.request = request
        self.cookies = cookies
        self._queued_messages = []
        self.used = False
        self.messages = []
        self.response_cookies = None
        self.load()  # load messages from cookies

    def load(self):
        """ Get messages from cookie """
        cookie_name = self.options['MESSAGES_COOKIE_NAME']
        try:
            data = self.request.cookies[cookie_name]
        except KeyError:
            data = None

        messages = None
        if data:
            messages = self._decode(data)

        if self.used:
            self.delete_cookie()

        self.messages = messages

        return messages

    def add_message(self, msg, msg_level=INFO):
        self._queued_messages.append(dict(msg=msg, lvl=msg_level))
        self.store_cookie()

    def info(self, msg):
        self.add_message(msg, msg_level=INFO)

    def debug(self, msg):
        self.add_message(msg, msg_level=DEBUG)

    def warning(self, msg):
        self.add_message(msg, msg_level=WARNING)

    def error(self, msg):
        self.add_message(msg, msg_level=ERROR)

    def success(self, msg):
        self.add_message(msg, msg_level=SUCCESS)

    def store_cookie(self):

        encoded_data = self._encode(self._queued_messages)

        expire_seconds = int(self.options['MESSAGES_COOKIE_EXPIRE'])
        if expire_seconds > 0:
            now = datetime.datetime.utcnow()
            expires_cookie = now + datetime.timedelta(seconds=expire_seconds)
        else:
            expires_cookie = None

        if not self.options['MESSAGES_COOKIE_DOMAIN'] or self.options['MESSAGES_COOKIE_DOMAIN'] == "":
            domain = self.request.host.split(':')[0]
        else:
            domain = self.options['MESSAGES_COOKIE_DOMAIN']

        cookies_options = {
            'HTTP_COOKIE_DOMAIN': domain,
            'HTTP_COOKIE_SECURE': self.options['MESSAGES_COOKIE_SECURE'],
            'HTTP_COOKIE_HTTPONLY': self.options['MESSAGES_COOKIE_HTTPONLY']
        }

        cookie_name = self.options['MESSAGES_COOKIE_NAME']

        cookies = self.cookies
        if not cookies:
            cookies = CookieManager(options=cookies_options)

        cookies[cookie_name] = encoded_data
        cookies[cookie_name].path = self.options['MESSAGES_COOKIE_PATH']
        cookies[cookie_name].expires = expires_cookie

        self.response_cookies = cookies

    def delete_cookie(self):

        cookie_name = self.options['MESSAGES_COOKIE_NAME']

        if not self.options['MESSAGES_COOKIE_DOMAIN'] or self.options['MESSAGES_COOKIE_DOMAIN'] == "":
            domain = self.request.host.split(':')[0]
        else:
            domain = self.options['MESSAGES_COOKIE_DOMAIN']

        cookies_options = {
            'HTTP_COOKIE_DOMAIN': domain,
            'HTTP_COOKIE_SECURE': self.options['MESSAGES_COOKIE_SECURE'],
            'HTTP_COOKIE_HTTPONLY': self.options['MESSAGES_COOKIE_HTTPONLY']
        }

        cookies = self.cookies
        if not cookies:
            cookies = CookieManager(options=cookies_options)

        del cookies[cookie_name]

        self.response_cookies = cookies

    def _encode(self, messages, encode_empty=False):

        if messages or encode_empty:
            value = signing_dumps(messages,
                                  key=self.options['MESSAGES_SECRET'])
            return value

    def _decode(self, data):

        if not data:
            return None

        try:
            messages = signing_loads(data,
                                     key=self.options['MESSAGES_SECRET'],
                                     max_age=self.options['MESSAGES_TIME_LIMIT'])
            self.used = True
        except:
            messages = None
            self.used = True

        return messages

    def render(self):

        xml_messages = list()
        if self.messages:
            for message in self.messages:

                xml = '''
                      <div class="alert alert-%(render_class)s">
                      <button type="button" class="close" data-dismiss="alert">&times;</button>
                      %(message)s
                      </div>
                      ''' % {'render_class': DEFAULT_TAGS[message['lvl']],
                             'message': message['msg']}
                xml_messages.append(xml)

            return '\n'.join(xml_messages)
        return ''


class ContextMessages(ContextManager):
    messages_manager = None

    def __init__(self, options=None):
        self.options = options
        messages_manager = MessagesManager(options)
        self.messages_manager = messages_manager

    def on_start(self, request):
        self.messages_manager.start(request)

    def add_message(self, msg, lvl=INFO):
        self.messages_manager.add_message(msg, lvl)

    def info(self, msg):
        self.messages_manager.add_message(msg, msg_level=INFO)

    def debug(self, msg):
        self.messages_manager.add_message(msg, msg_level=DEBUG)

    def warning(self, msg):
        self.messages_manager.add_message(msg, msg_level=WARNING)

    def error(self, msg):
        self.messages_manager.add_message(msg, msg_level=ERROR)

    def success(self, msg):
        self.messages_manager.add_message(msg, msg_level=SUCCESS)

    def render(self):
        return self.messages_manager.render()

    @property
    def response_cookies(self):
        return self.messages_manager.response_cookies
